// Emacs style mode select   -*- C++ -*-
//-----------------------------------------------------------------------------
//
// Copyright(C) 2002 James Haley, Simon Howard, Florian Schulze
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
//----------------------------------------------------------------------------
//
// DESCRIPTION:
//
// Windows MCI MIDI Interface
//
//----------------------------------------------------------------------------

#include <stdlib.h>
#include <stdio.h>
#include <limits.h>

#include <sdl.h>
#include <sdl_syswm.h>
#include "mmus2mid.h"

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <mmsystem.h>

int MusicLoaded = 0;
int MusicLoop = 0;
MIDI mididata;

void usermsg(char *, ...);

static int myMCIErrNum = 0;

//
// SafemciSendString
//
// Wrapper function for mciSendString, with error detection
//
char *SafemciSendString(char *cmd)
{
   static char errora[1024];
   int err;

   memset(errora, 0, 1024);
   myMCIErrNum = 0;
   
   err = mciSendString(cmd, errora, 1023, NULL);
   if(err)
   {
      mciGetErrorString(err, errora, 1024);
      usermsg("%s\n", errora);
      myMCIErrNum = err;
   }
   return errora;
}

//
// SafemciSendStringEvt
//
// Wrapper function for mciSendString, with error detection
// This version takes an HWND to which MCI can send notification
// messages
//
char *SafemciSendStringEvt(char *cmd, HWND hWnd)
{
   static char errora[1024];
   int err;

   memset(errora, 0, 1024);
   myMCIErrNum = 0;
   
   err = mciSendString(cmd, errora, 1023, hWnd);
   if(err)
   {
      mciGetErrorString(err, errora, 1024);
      usermsg("%s\n", errora);
      myMCIErrNum = err;
   }
   return errora;
}

// haleyjd: keep this here in case more initialization is necessary 
// in the future

void Win_InitMusic(void)
{
   printf("I_InitMusic: Using MCI-MIDI Device\n");
}

//
// Returns an HWND pointing to Eternity's SDL window, needed to
// capture MCI events
//
HWND Win_GetSDLWmInfo(void)
{
   // get a handle to the SDL window
   SDL_SysWMinfo wmInfo;
   SDL_VERSION(&(wmInfo.version));
   SDL_GetWMInfo(&wmInfo);
   
   // return the HWND
   return wmInfo.window;
}

//
// Win_HandleStopEvent
//
// Called by I_GetEvent in i_video.c -- dispatches the MCI event
// scheduled to occur when music has stopped playing
//
void Win_HandleStopEvent(SDL_Event *event)
{
   UINT msg;

   if(!MusicLoaded)
      return;

   msg = event->syswm.msg->msg;

   if(msg != MM_MCINOTIFY)
      return;

   if(MusicLoop)
   {
      HWND windowHandle = Win_GetSDLWmInfo();

      if(_stricmp(SafemciSendString("status doommusic mode"),"stopped") == 0)
         SafemciSendStringEvt("play doommusic from 0 notify", windowHandle);
   }
}

void Win_PlaySong(int handle, int looping)
{
   HWND windowHandle = Win_GetSDLWmInfo();

   if((handle >= 0) && MusicLoaded)
   {
      if(_stricmp(SafemciSendString("status doommusic mode"),"playing") == 0)
         return;      
      SafemciSendStringEvt("play doommusic from 0 notify", windowHandle);
      MusicLoop = looping;
   }
}

void Win_PauseSong(int handle)
{
   /*
   if((handle >= 0) && MusicLoaded)
      SafemciSendString("pause doommusic");
   */
}

void Win_ResumeSong(int handle)
{
   /*
   HWND windowHandle = Win_GetSDLWmInfo();

   if((handle >= 0) && MusicLoaded)
      SafemciSendStringEvt("resume doommusic notify", windowHandle);
   */
}

void Win_StopSong(int handle)
{
   if((handle >= 0) && MusicLoaded)
      SafemciSendString("stop doommusic");
}

void Win_UnRegisterSong(void)
{
   MusicLoaded = 0;
   SafemciSendString("close doommusic");
}

// note: not the real prototype, but boolean causes type conflicts
int M_WriteFile(const char *, void *, int);

int Win_RegisterSong(void)
{  
   UBYTE *mid;
   int midlen;

   MIDIToMidi(&mididata, &mid, &midlen);   
   if(M_WriteFile("eetemp.mid", mid, midlen))
   {
      SafemciSendString("open eetemp.mid type sequencer alias doommusic");
      if(!myMCIErrNum)
         MusicLoaded = 1;
   }
   free(mid);

   return 0;
}

// EOF